; 
; Wasteland IDE fix by Grue
; 
; 	This fix is based on the Wasteland 1MB REU version, which was posted in Lemon64 forums in 2013 by user mood_swing.
;	He has since sadly removed all postings and links to his fix.
;	In the end, I had to write much more code and figure out strange behavior due to my inexperience on how
;	loading works on C64. Problematic is also that the game uses almost all the available memory
;	if you see artifacts on screen during filesave it just means that game is saving under IO memory, data is buffered
; 	temporarily on bitmap on such case.
;	I hope this source code will help to fix more games for IDE64, 
;	
;	Thanks to:
; 	Soci / Singular for help and code!
;	TNT / Beyond Force for help, support and teaching me 6502 asm.
;	Trurl / Extend for help, test and suggestions
;	All IDE64 fans!
; 
;	Tools used for making this IDE64 fix possible, not in any special order
;
;	IDE64
;	64tass
;	010 Editor
; 	Ida pro
;	Ultimate 64 + ucodenet
;	C64debugger
;	Vice
;	Sublime text 3
;	Max roast level Coffee


lindicator		= $6000+(8*39)				; Memory location for the indicator character
savebuffer		= $6000						; savebuffer for saving under IO

track			= $f8
sector			= $f9
ldr_cmd			= $f3
load_until		= $fd
load_adr_lo		= $fe
load_adr_hi		= $ff
open			= $ffc0
read			= $def4
write			= $def1
setnam			= $ffbd
setlfs			= $ffba

				*= $c000
				lda #12									; file number
				ldx $ba									; current drive
				tay
				jsr setlfs								; setup file for opening
				lda #4 									; lenght of the filename
				ldx #<filename							; location of the filename to open
				ldy #>filename
				jsr setnam								; setup filename and name lenght
				jsr open								; Open file

				lda #15									; Setup Command channel to seek_to_position
				ldx $ba									; Current drive
				tay										; Secondary address
				jsr setlfs
				lda #0
				jsr setnam
				jsr open								; open cmd channel
				ldx #30-1								; swap vectorspace
-				lda $0259,x
				sta filetables-$fc00+$c300,x
				dex
				bpl -
				ldx #$54-1 								; -1 since we use dex bpl method and zero counts as well
-				lda $70,x
				sta zp_swap-$fe2c+$c52c,x
				dex
				bpl -

				ldx #1									; Copy kernal jump locations from the vectors
-				lda $31e,x								; So we dont need to care if they get 
				sta chkin_+1-loader_adr+$c42c,x			; overwritten by the game load (they do...)
				lda $320,x								
				sta chkout_+1-loader_adr+$c42c,x
				lda $322,x
				sta clrchn_+1-loader_adr+$c42c,x
				lda $326,x
				sta chrout_+1-loader_adr+$c42c,x
				dex
				bpl -

				sei										; Disable interrupts as we switch kernal rom off
				lda #$35
				sta $01

				lda #$7f
				sta $dc0d 								; set IRQ interrupts
				sta $dd0d								; set NMI interrupts
				lda $dc0d 								; clear them by reading
				lda $dd0d
				ldx #0									; set Screen + Color ram for bitmap gfx
-				lda #1
				sta $d800,x								; color ram location
				sta $d900,x
				sta $da00,x
				sta $db00,x
				lda #$6f
				sta $5c00,x								; screen ram location game uses
				sta $5d00,x
				sta $5e00,x
				sta $5f00,x
				cpx #$2e
				bcs +
				lda vic2_tbl,x 							; setup VIC-II registers to predefined values
				sta $d000,x
				cpx #$1d
				bcs +
				lda #0
				sta $d400,x
+				inx
				bne -

				lda $dd00 								; set VIC-II bank to $4000-$7fff
				and #$fc
				ora #2
				sta $dd00
				lda #$3f 								
				sta $dd02

				lda #$2f								; setup cpu data direction registers to defaults
				sta 0
				ldx #$ff
				txs 									; set stackpointer to default value
				
				ldx #4 									; transfer loader code into its place $fc00-$ffff
				ldy #0 									; this will also set NMI & IRQ addresses
mod:			lda loader_code,y
				sta $fc00,y
				iny
				bne mod
				inc mod+2
				inc mod+5
				dex
				bne mod

				ldy #0 									; copy block of game data into its place
-				lda game_dat_tbl,y
				sta $5b00,y
				iny
				bne -

				lda #$34 								; switch IO addresses off for RAM under IO
				sta $1
				ldy #0
-				lda io_code_tbl,y 						; Copy code under IO space
				sta $dd80,y
				iny
				bpl -

				lda #$35 								; IO back on, Kernal off
				sta 1
				lda #0
				sta load_adr_lo
				lda #$c6
				sta load_adr_hi 
				clc
				adc #3 									; blocks
				sta load_until 							; We want to load 3 blocks, untill $c9ff
				lda #1 									; 1=Load
				ldy #$22								; Track 34
				ldx #$e 								; Sector 14
				jsr Loader_Call							; Load $c600-$c9ff

				lda #0
				sta load_adr_lo
				lda #2
				sta load_adr_hi
				clc
				adc #$2e 								;load $2e blocks
				sta load_until
				lda #1
				ldy #$22 								; Start transfer from Track 34
				ldx #$a 								; Sector 10 and continue there with interleave tables
				jsr Loader_Call							; Load $0200-$2fff

				lda #0
				sta load_adr_lo
				lda #$7e
				sta load_adr_hi
				clc
				adc #$f 								; get 15 blocks
				sta load_until
				lda #1
				ldy #4 									; Start Track 4
				ldx #$10 								; Sector 16
				jsr Loader_Call							; Load $7e00-$8dff
				lda #1									; enable IRQ
				sta $d01a
				jmp $7e00								; Start the game


vic2_tbl:		.byte   0,  0,  0,  0,  0,  0,  0,  0; $0
		        .byte   0,  0,  0,  0,  0,  0,  0,  0; $8
		        .byte   0,  0,  0,  0,  0,  0,$18,  0; $10 ; $d017 0001 1000 = multicolor mode, 40 cols
		        .byte $78,  0,  0,  0,  0,  0,  0,  0; $18 ; $d018 0111 1000 = sceeen $5c00, bitmap $6000 
		        .byte   0,  0,  2,  3,  4,  0,  0,  0; $20 ; $d022, 23, 24   = bitmap colors red, cyan, purple
		        .byte   0,  0,  0,  0,  0,  0,  0	 ; $28

game_dat_tbl:   .byte   2,  2,  2,  2,  2,  0,  0,  2; 0
		        .byte   0,  2,  2,  0,  0,  1,  2,  0; 8
		        .byte   2,  2,  1,  0,  0,  2,  0,  0; $10
		        .byte   1,  1,  0,  1,  0,  1,  1,  0; $18
		        .byte   0,  2,  2,  2,  2,  2,  0,  2; $20
		        .byte   0,  2,  2,  0,  0,  0,  0,  0; $28
			    .byte   0,  2,  0,  0,  0,  0,  2,  0; $30
			    .byte   0,  1,  1,  0,  0,  2,  2,  0; $38
			    .byte   2,  0,  2,  2,  0,  0,  2,  2; $40
			    .byte   0,  2,  2,  2,  0,  2,  1,  0; $48
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $50
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $58
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $60
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $68
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $70
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $78
			    .byte   0,  0,  0,  0,  0,  0,  0,  0; $80
			    .byte   0,  0,  0,  0,  0,  0,  0,$ff; $88
			    .byte $ff,$ff,$ff,  1,  1,  1,  3,  4; $90
			    .byte   1,  3,  4,  3,  4,  1,  3,  4; $98
			    .byte   0,  0,  0,  0,$f1,$f3,$f5,$f7; $a0
			    .byte   0,  0,  0,$55,$73,$65,$20,$45; $a8
			    .byte $6e,$63,$20,$4f,$72,$64,$65,$72; $b0
			    .byte $20,$44,$69,$73,$62,$61,$6e,$64; $b8
			    .byte $20,$56,$69,$65,$77,$20,$53,$61; $c0
			    .byte $76,$65,$20,$52,$61,$64,$69,$6f; $c8
			    .byte $20,$20,$20,  0,$49,$4b,$4a,$4c; $d0
			    .byte $1b,$20,  0,  0,  0,  0,  0,  0; $d8
			    .byte   0,  0,  0,  5,  0,  6,  2,  4; $e0
			    .byte   3,  7,  1,  8,  0,  1,  2,  3; $e8
			    .byte   4,  5,  6,  7,  8,  0,  6,  3; $f0
			    .byte   4,  2,  1,  1,  1,  1,  1	 ; $f8
io_code_tbl:	.byte $a0,  0,$a2,  0,$b9,$ff,$ff,$9d; 0
			    .byte $ff,$ff,$e8,$c8,$c0,$20,$d0,$f4; 8
			    .byte $a0,  0,$a2,$20,$20,$9d,$dd,$e6; $10
			    .byte $13,$e6,$15,$a2,  2,$b1,$12,$91; $18
			    .byte $14,$c8,$b1,$12,$91,$14,$c8,$b1; $20
			    .byte $12,$91,$14,$c8,$b1,$12,$91,$14; $28
			    .byte $c8,$b1,$12,$91,$14,$c8,$b1,$12; $30
			    .byte $91,$14,$c8,$b1,$12,$91,$14,$c8; $38
			    .byte $b1,$12,$91,$14,$c8,$ca,$d0,$d5; $40
			    .byte $60,$8c,$91,$dd,$8e,$97,$dd,$8e; $48
			    .byte $99,$dd,$a0,  0,$a2,  8,$99,$a1; $50
			    .byte $dd,$48,$98,$18,$69,  5,$a8,$68; $58
			    .byte $ca,$d0,$f3,$60,  3,$22,  1,$c8; $60
			    .byte $88,$c8,$24,  1,$25,  1,$24,  1; $68
			    .byte $e8,$ca,$e8,  8,$28,  0,  0,$18; $70
			    .byte   8,  0,  0,  0,  0,  0,  0,  0; $78
			    .byte   0,  0,  0,  0,  0,  0; $80

filename 		.text "WL,M"

				*=$c300
loader_code:
				
                .logical $fc00
                jmp Loader_Call
save_table:   	.byte   0,  0,  0,  0,  0,  0,  0,  0; 0
                .byte   0,  0,  0,  0,  0,  0,  0,$ff; 8
                .byte $ff,  0,$ff,$ff,  0,  0,  0,  0; $10
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $18
                .byte   0,  0,$ff,  0,  0,  0,  0,  0; $20
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $28
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $30
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $38
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $40
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $48
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $50
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $58
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $60
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $68
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $70
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $78
                .byte   0,  0,  0,  0,  0,  0,  0,  0; $80
                .byte   0,  0,  0,  0, $f, $f,  0,  0; $88
                .byte  $f, $f, $f, $f, $f, $f,  0,  0; $90
                .here
                *=$c390
     			.logical $fc90
     			
restore_save_tbl:	
				lda #$65					; Load block from predefined $00056500 address from wl,reu to $fc03
				sta seekloc+1
				lda #$5
				sta seekloc+2
				lda #$8c 					; Modify load code to get $8c bytes instead of default 1 block
				sta savehack+1 
				lda #$00
				sta savehack+3
				lda $fe 					; Store original save ram location
				ldy $ff
				sta save_adr
				sty save_adr+1
				lda #$fc
				sta $ff
				lda #$03
				sta $fe
				jsr read_from_file			
				ldx #$01 					; Restore original 1 block load
				stx savehack+3
				dex
				stx savehack+1
				lda save_adr 				; Restore original save ram location
				ldy save_adr+1
				sta $fe
				sty $ff
				rts
disktable:		.byte $60,$24,$1f,$19,$12 				; First value is bogus, we just have 4 disk images
intl_tbl2:		.byte   0,  8,$10,  5, $d,  2, $a,$12	; 19 sectors on speedzone
				.byte   7, $f,  4, $c,  1,  9,$11,  6
				.byte  $e,  3, $b,  0

filetables		.fill 30,$00
				
.here	
				*=$c40b
				.logical $fd0b

Step_and_return_max_sector_on_track:

				dey                     		; current sector on y
				bpl ++				     		; if sector positive exit
				dex                     		; decrement track number
				cpx #$12
				bne +			 			    ; if track 18, skip
				dex                     		; if track was 18, proceed to track 17
+               lda $c930,x 	        		; read max sectors on track
				tay                     		; return sector on y
+				rts
				ldx $11 						; these locations are referenced in original game code.
				ldy $10
				lda #0
				pha
				lda #0
				sta $fe
				lda #$5a
				sta $ff
				pla
				jmp $03aa 						; vector_load_data
.here

				*=$c42c
				.logical $fd2c
loader_adr:				
seekcmd			.byte $50,12		; P cmd + secondary address
seekloc			.byte 0,0,0,0,13	; cmd continues with seek address, bits 0-7, 8-15, 15-23 and 24-31 in this order

read_from_file:	lda #$60
				sta chra+1 						; Modify adress for letter L
				jsr indicator					; draw L
				jsr smol						; we are using less than 1MB file, we need to tune seek address.
				jsr swap_memory2 					; swap zp-locations
				jsr seek_to_position			; find position to read from and open needed files
				ldx #12							; file number
chkin_			jsr $ffff						; set for input
				jsr special_selfoverwrite 		; check if we are loading on $200 and handle it
				lda #$fe 						; c64 address to load to
savehack:		ldx #0						 
				ldy #1							; set to load 1 block
				jsr read
				jsr special_selfoverwrite
end:			jsr clrchn_
				
swap_memory2:	jsr swap_memory
				ldx #$54-1
s2:				ldy zp_swap,x
				lda $70,x
				sta zp_swap,x
				sty $70,x
				dex
				bpl s2
				rts

special_selfoverwrite:
				lda $ff 						; check if we are loading on ourselves
				cmp #2  						; at $0200 page
				bne + 							; if not, skip
				jsr swap_memory						; if we are then restore healthy values
+				rts

swap_memory:	ldx #30-1						; swap filetables
-				ldy	filetables,x
				lda $0259,x
				sta	filetables,x
				tya
				sta	$0259,x
				dex
				bpl -
				rts

chkout_			jmp $ffff
seek_to_position:
				ldx #15
				jsr chkout_ 
				ldx #<-7 						; lenght of the seek cmd string
-				lda seekcmd-(<-7),x 			; send it to cmd channel #15
chrout_			jsr $ffff 						; using chrout
				inx
				bne -
clrchn_			jmp $ffff 						; call clrchn

write_to_file: 	lda #$98						; Modify adress for letter S
				sta chra+1
				jsr indicator					; Draw S
				jsr smol 						; we are using smaller image file, we need to tune seek address.
				jsr swap_memory2				; swap memory
				jsr check 						; check if we're gonna save under IO area
				jsr seek_to_position 			; seek file position for save
				ldx #12							; file number
				jsr chkout_						; set for input
				lda #$fe 						; c64 address to load to
				ldx #0						 
				ldy #1							; set to save 1 block
				jsr write
				jsr clrchn_
				jsr seek_to_position			; flush written data to disk
				jmp end

check:			lda $ff
				and #$f0
				cmp #$d0 						; check if we are trying to save under $d000-$dfff
				bne +	 						; nope we are not						
buffercopy:		dec 1							; if we are, copy memory and save from buffer
				ldy #$00
-				lda ($fe),y
				sta savebuffer,y
				iny
				bne -
				lda #<savebuffer
				sta $fe
				lda #>savebuffer
				sta $ff
				inc $01
+				rts

smol:			sec								; do the math, substract #4 from the seek value
				lda seekloc+2					; I learned here the hard way that you ALWAYS
				sbc #4							; need to set C before substract
				sta seekloc+2
				rts

indicator		ldx #7 							; print character to screen
chra:			lda $c760,x
				sta lindicator,x
				dex
				bpl chra
				rts
grapbitmap		ldx #7 							; get backup of the bitmap
-				lda lindicator,x
				sta memsave,x
				dex
				bpl -
				rts
restbitmap		ldx #7 							; restore bitmap
-				lda memsave,x
				sta lindicator,x
				dex
				bpl -
				rts				
memsave:		.byte 0,0,0,0,0,0,0,0
				.here

				
				*=$c52c
				.logical $fe2c

save_adr:		.word $0000
zp_swap:		.fill 84, 00
								
Load:			jsr Find_Load_Addr	
				jsr read_from_file						; call Loader
				ldx ldr_cmd
				dex
				bne Exit_Loader							; if not 0 gtfo
				lda $ff
				cmp $fd
				beq Exit_Loader							; all blocks done, exit
				inc $ff
				ldx track
				ldy sector
				jsr Step_and_return_max_sector_on_track	; find next track/sector
				stx track
				sty sector
				jmp Load 								; check if theres more

Save: 			jsr restore_save_tbl               		; Load $8c bytes to filesave_tbl
				clc 									; clear carry as we do math
				ldx floppy_side							; value 1-4
				lda filesave_tbl,x 						; $23,$46,$69,$20
				adc track 								; + track (max value $8c)
				tax
				lda #$ff
				sta save_table-1,x						; mark $ff on savetable, theres no track 0 so -1 from table index
				jsr Find_Load_Addr 						; find Sector to save to
				jsr write_to_file 						; Finally do write

Exit_Loader:   	jsr restbitmap 							; clear load/save indicator
				pla
				sta 1
				lda #0
				sta $fc
				cmp #1
				rts
				.here				

				*=$c5fc
				.logical $fefc
				jmp $2967 						; have to be here, do not break original game code
				.here

				*=$c5ff
				.logical $feff					; game keeps disk side info here
floppy_side	    .byte $00
				
Loader_Call:    sta ldr_cmd 					; store registers containing needed info
                stx sector
                sty track
                lda 1 							; store original value of $01 so we can restore it later
                pha
                lda #$35 						; Switch kernal and basic rom off (kernal already was, since we are here)
                sta 1
                jsr grapbitmap 					; make copy of bitmap area where L/S is plotted
                lda ldr_cmd
                cmp #2 							; check if we got called for save
                beq Save
                bcs Exit_Loader 				; if command is something else than 1 or 2, gtfo
                jmp Load 						; Since cmd must be 1, jump to Load routine

Find_Load_Addr: 								; this routine seeks position in the wl game file
				ldy track						; its not coded by me. 
				dey								; it is 1:1 the same on REU version including 
				tya								; tables
				asl
				tay
				lda floppy_side
				asl
				tax
				lda fileaddr_tbl2,x
				adc fileaddr_tbl,y
				sta seekloc+1
				lda fileaddr_tbl2+1,x
				adc fileaddr_tbl+1,y
				pha
				jsr interleave
				adc seekloc+1
				sta seekloc+1
				pla
				adc #0
				sta seekloc+2;
				rts

fileaddr_tbl:	.byte   0,  0,$15,  0,$2a,  0,$3f,  0
				.byte $54,  0,$69,  0,$7e,  0,$93,  0
				.byte $a8,  0,$bd,  0,$d2,  0,$e7,  0
				.byte $fc,  0,$11,  1,$26,  1,$3b,  1
				.byte $50,  1,$65,  1,$78,  1,$8b,  1
				.byte $9e,  1,$b1,  1,$c4,  1,$d7,  1
				.byte $ea,  1,$fc,  1, $e,  2,$20,  2
				.byte $32,  2,$44,  2,$56,  2,$67,  2
				.byte $78,  2,$89,  2,$9a,  2
fileaddr_tbl2:	.byte   0,  4,$ab,  6,$56,  9,  1, $c
filesave_tbl:	.byte   0,$23,$46,$69,$20


interleave:		lda track 								; load track
				ldx #4 									; We have 4 different track lenghts
-				cmp disktable,x 						; Check if current track fits into track lenght
				dex 									; move to index from 0 vs 1 (tables 0-3, not1-4)
				bcs -	 								; if track was > zone, check next
				txa 									; we finally reached table that track no fits into		
				asl                                     ; multiply x 2
				tax 									; now we know where to look correct interleave table address
				lda intl_tbl_loc,x 						; get memory location for table and set it
				sta intl_read+1
				lda intl_tbl_loc+1,x
				sta intl_read+2
				ldy sector 								; what sector we are on
intl_read:		lda intl_tbl0,y 						; read from the interleave table sector number for track
				rts 									; return with correct sector in Y
		
intl_tbl_loc:	.word intl_tbl0, intl_tbl1, intl_tbl2, intl_tbl3

intl_tbl0:		.byte   0,  6, $c,  1,  7, $d,  2,  8 ; 17 sectors on speedzone
				.byte  $e,  3,  9, $f,  4, $a,$10,  5
				.byte  $b,  0
intl_tbl1:		.byte   0,  7, $e,  3, $a,$11,  6, $d ; 18 sectors on speedzone
				.byte   2,  9,$10,  5, $c,  1,  8, $f
				.byte   4, $b, 0
intl_tbl3:		.byte   0, $a,$14,  9,$13,  8,$12,  7 ; 21 sectors on speedzone
				.byte $11,  6,$10,  5, $f,  4, $e,  3
				.byte  $d,  2, $c,  1, $b,  0

				.here
				*=$c6fa
				.word $04f6, $0000, $04f9
